@@ -2,16 +2,17 @@
 
 #  ************************** Copyrights and license ***************************
 #
-# This file is part of gcovr 5.1, a parsing and reporting tool for gcov.
+# This file is part of gcovr 6.0, a parsing and reporting tool for gcov.
 # https://gcovr.com/en/stable
 #
 # _____________________________________________________________________________
 #
-# Copyright (c) 2013-2022 the gcovr authors
+# Copyright (c) 2013-2023 the gcovr authors
 # Copyright (c) 2013 Sandia Corporation.
-# This software is distributed under the BSD License.
 # Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 # the U.S. Government retains certain rights in this software.
+#
+# This software is distributed under the 3-clause BSD License.
 # For more information, see the README.rst file.
 #
 # ****************************************************************************
@@ -20,6 +21,7 @@ from __future__ import annotations
 from argparse import ArgumentParser, ArgumentTypeError, SUPPRESS
 from inspect import isclass
 from locale import getpreferredencoding
+import logging
 from multiprocessing import cpu_count
 from typing import Iterable, Any, List, Optional, Union, Callable, TextIO, Dict
 from dataclasses import dataclass
@@ -27,9 +29,11 @@ import datetime
 import os
 import re
 
-from .utils import FilterOption
+from .utils import FilterOption, force_unix_separator
 from .writer.html import CssRenderer
 
+logger = logging.getLogger("gcovr")
+
 
 def check_percentage(value: str) -> float:
     r"""
@@ -68,6 +72,22 @@ def check_input_file(value: str, basedir: str = None) -> str:
     return os.path.abspath(value)
 
 
+def relative_path(value: str, basedir: str = None) -> str:
+    r"""
+    Make a absolute path if value is a relative path.
+    """
+    if not value:
+        raise ArgumentTypeError("Should not be set to an empty string.") from None
+
+    if basedir is None:
+        basedir = os.getcwd()
+
+    if not os.path.isabs(value):
+        value = os.path.join(basedir, value)
+    value = os.path.normpath(value)
+    return os.path.relpath(value, os.getcwd())
+
+
 def timestamp(value: str) -> datetime.datetime:
     from .timestamps import parse_timestamp  # lazy import
 
@@ -77,6 +97,50 @@ def timestamp(value: str) -> datetime.datetime:
         raise ArgumentTypeError(f"{ex}: {value!r}") from None
 
 
+def source_date_epoch() -> Optional[datetime.datetime]:
+    """
+    Load time from SOURCE_DATE_EPOCH, if it exists.
+    See: <https://reproducible-builds.org/docs/source-date-epoch/>
+
+    Examples:
+    >>> monkeypatch = getfixture("monkeypatch")
+    >>> caplog = getfixture("caplog")
+
+    Example: can be empty
+    >>> with monkeypatch.context() as mp:
+    ...   mp.delenv("SOURCE_DATE_EPOCH", raising=False)
+    ...   print(source_date_epoch())
+    None
+
+    Example: can contain timestamp
+    >>> with monkeypatch.context() as mp:
+    ...   mp.setenv("SOURCE_DATE_EPOCH", "1677067226")
+    ...   print(source_date_epoch())
+    2023-02-22 12:00:26+00:00
+
+    Example: can contain invalid timestamp
+    >>> with monkeypatch.context() as mp:
+    ...   mp.setenv("SOURCE_DATE_EPOCH", "not a timestamp")
+    ...   print(source_date_epoch())
+    None
+    >>> for m in caplog.messages: print(m)
+    Ignoring invalid environment variable SOURCE_DATE_EPOCH='not a timestamp'
+    """
+
+    ts = os.environ.get("SOURCE_DATE_EPOCH")
+
+    if ts:
+        try:
+            return datetime.datetime.fromtimestamp(int(ts), datetime.timezone.utc)
+        except Exception:
+            logger.warning(
+                "Ignoring invalid environment variable SOURCE_DATE_EPOCH=%r",
+                ts,
+            )
+
+    return None
+
+
 class OutputOrDefault:
     """An output path that may be empty.
 
@@ -477,6 +541,12 @@ def _get_value_from_config_entry(
         except (ValueError, ArgumentTypeError) as err:
             raise cfg_entry.error(str(err))
 
+    elif option.config == "add-tracefile":  # Special case for patterns
+        assert (
+            cfg_entry.filename is not None
+        ), "conversion function must derive base directory from filename"
+        basedir = os.path.dirname(cfg_entry.filename)
+        value = os.path.join(basedir, cfg_entry.value)
     else:
         value = cfg_entry.value
 
@@ -510,6 +580,9 @@ def _get_converter_function(
     if option_type is check_input_file:
         return lambda value: check_input_file(value, basedir)
 
+    if option_type is relative_path:
+        return lambda value: relative_path(value, basedir)
+
     if option_type is OutputOrDefault:
         return lambda value: OutputOrDefault(value, basedir)
 
@@ -522,12 +595,7 @@ def _assign_value_to_dict(
     option: GcovrConfigOption,
     is_single_value: bool,
 ) -> None:
-
-    if option.action in ("store", "store_const"):
-        namespace[option.name] = value
-        return
-
-    if option.action == "append":
+    if option.action == "append" or option.nargs == "*":
         append_target = namespace.setdefault(option.name, [])
         if is_single_value:
             append_target.append(value)
@@ -535,6 +603,10 @@ def _assign_value_to_dict(
             append_target.extend(value)
         return
 
+    if option.action in ("store", "store_const"):
+        namespace[option.name] = value
+        return
+
     assert False, f"unexpected action for {option.name}: {option.action!r}"
 
 
@@ -626,6 +698,7 @@ GCOVR_CONFIG_OPTIONS = [
             "The --root is the default --filter."
         ),
         default=".",
+        type=relative_path,
     ),
     GcovrConfigOption(
         "add_tracefile",
@@ -654,6 +727,7 @@ GCOVR_CONFIG_OPTIONS = [
             "Search these directories for coverage files. "
             "Defaults to --root and --object-directory."
         ),
+        type=relative_path,
     ),
     GcovrConfigOption(
         "config",
@@ -663,6 +737,7 @@ GCOVR_CONFIG_OPTIONS = [
             "Load that configuration file. "
             "Defaults to gcovr.cfg in the --root directory."
         ),
+        type=relative_path,
     ),
     GcovrConfigOption(
         "respect_exclusion_markers",
@@ -735,6 +810,13 @@ GCOVR_CONFIG_OPTIONS = [
         help="Report the decision coverage. For HTML and JSON report.",
         action="store_true",
     ),
+    GcovrConfigOption(
+        "exclude_calls",
+        ["--calls"],
+        group="output_options",
+        help="Report the calls coverage. For HTML report.",
+        action="store_false",
+    ),
     GcovrConfigOption(
         "sort_uncovered",
         ["-u", "--sort-uncovered"],
@@ -808,7 +890,7 @@ GCOVR_CONFIG_OPTIONS = [
         metavar="OUTPUT",
         help=(
             "Add annotated source code reports to the HTML report. "
-            "Implies --html. "
+            "Implies --html, can not be used together with --html-nested. "
             "OUTPUT is optional and defaults to --output."
         ),
         nargs="?",
@@ -817,10 +899,27 @@ GCOVR_CONFIG_OPTIONS = [
         const=OutputOrDefault(None),
     ),
     GcovrConfigOption(
-        "html_details_syntax_highlighting",
-        ["--html-details-syntax-highlighting"],
+        "html_nested",
+        ["--html-nested"],
         group="output_options",
-        help="Use syntax highlighting in HTML details page. Enabled by default.",
+        metavar="OUTPUT",
+        help=(
+            "Add annotated source code reports to the HTML report. "
+            "A page is created for each directory that summarize subdirectories "
+            "with aggregated statistics. "
+            "Implies --html, can not be used together with --html-details. "
+            "OUTPUT is optional and defaults to --output."
+        ),
+        nargs="?",
+        type=OutputOrDefault,
+        default=None,
+        const=OutputOrDefault(None),
+    ),
+    GcovrConfigOption(
+        "html_syntax_highlighting",
+        ["--html-syntax-highlighting", "--html-details-syntax-highlighting"],
+        group="output_options",
+        help="Use syntax highlighting in HTML source page. Enabled by default.",
         action="store_const",
         default=True,
         const=True,
@@ -887,6 +986,60 @@ GCOVR_CONFIG_OPTIONS = [
         ),
         default=90.0,
     ),
+    GcovrConfigOption(
+        "html_medium_threshold_branch",
+        ["--html-medium-threshold-branch"],
+        group="output_options",
+        metavar="MEDIUM_BRANCH",
+        type=check_percentage,
+        help="If the coverage is below MEDIUM_BRANCH, the value is marked "
+        "as low coverage in the HTML report. "
+        "MEDIUM_BRANCH has to be lower than or equal to value of --html-high-threshold-branch "
+        "and greater than 0. "
+        "If MEDIUM_BRANCH is equal to value of --html-medium-threshold-branch the report has "
+        "only high and low coverage. Default is taken from --html-medium-threshold.",
+        default=None,
+    ),
+    GcovrConfigOption(
+        "html_high_threshold_branch",
+        ["--html-high-threshold-branch"],
+        group="output_options",
+        type=check_percentage,
+        metavar="HIGH_BRANCH",
+        help="If the coverage is below HIGH_BRANCH, the value is marked "
+        "as medium coverage in the HTML report. "
+        "HIGH_BRANCH has to be greater than or equal to value of --html-medium-threshold-branch. "
+        "If HIGH_BRANCH is equal to value of --html-medium-threshold-branch the report has "
+        "only high and low coverage. Default is taken from --html-high-threshold.",
+        default=None,
+    ),
+    GcovrConfigOption(
+        "html_medium_threshold_line",
+        ["--html-medium-threshold-line"],
+        group="output_options",
+        metavar="MEDIUM_LINE",
+        type=check_percentage,
+        help="If the coverage is below MEDIUM_LINE, the value is marked "
+        "as low coverage in the HTML report. "
+        "MEDIUM_LINE has to be lower than or equal to value of --html-high-threshold-line "
+        "and greater than 0. "
+        "If MEDIUM_LINE is equal to value of --html-medium-threshold-line the report has "
+        "only high and low coverage. Default is taken from --html-medium-threshold.",
+        default=None,
+    ),
+    GcovrConfigOption(
+        "html_high_threshold_line",
+        ["--html-high-threshold-line"],
+        group="output_options",
+        type=check_percentage,
+        metavar="HIGH_LINE",
+        help="If the coverage is below HIGH_LINE, the value is marked "
+        "as medium coverage in the HTML report. "
+        "HIGH_LINE has to be greater than or equal to value of --html-medium-threshold-line. "
+        "If HIGH_LINE is equal to value of --html-medium-threshold-line the report has "
+        "only high and low coverage. Default is taken from --html-high-threshold.",
+        default=None,
+    ),
     GcovrConfigOption(
         "html_tab_size",
         ["--html-tab-size"],
@@ -999,6 +1152,15 @@ GCOVR_CONFIG_OPTIONS = [
         ),
         action="store_true",
     ),
+    GcovrConfigOption(
+        "json_base",
+        ["--json-base"],
+        group="output_options",
+        metavar="PATH",
+        help="Prepend the given path to all file paths in JSON report.",
+        type=lambda p: force_unix_separator(os.path.normpath(p)),
+        default=None,
+    ),
     GcovrConfigOption(
         "csv",
         ["--csv"],
@@ -1045,10 +1207,12 @@ GCOVR_CONFIG_OPTIONS = [
             "Override current time for reproducible reports. "
             "Can use `YYYY-MM-DD hh:mm:ss` or epoch notation. "
             "Used by HTML, Coveralls, and Cobertura reports. "
-            "Default: current time."
+            "Default: Environment variable SOURCE_DATE_EPOCH "
+            "(see https://reproducible-builds.org/docs/source-date-epoch) "
+            "or current time."
         ),
         type=timestamp,
-        default=datetime.datetime.now(),
+        default=source_date_epoch() or datetime.datetime.now(),
     ),
     GcovrConfigOption(
         "filter",
@@ -1153,9 +1317,36 @@ GCOVR_CONFIG_OPTIONS = [
         "exclude_function_lines",
         ["--exclude-function-lines"],
         group="gcov_options",
-        help="Exclude coverage from lines defining a function Default: {default!s}.",
+        help="Exclude coverage from lines defining a function. Default: {default!s}.",
         action="store_true",
     ),
+    GcovrConfigOption(
+        "merge_mode_functions",
+        ["--merge-mode-functions"],
+        metavar="MERGE_MODE",
+        group="gcov_options",
+        choices=[
+            "strict",
+            "merge-use-line-0",
+            "merge-use-line-min",
+            "merge-use-line-max",
+            "separate",
+        ],
+        default="strict",
+        help=(
+            "The merge mode for functions coverage from different gcov files for same sourcefile."
+            "Default: {default!s}."
+        ),
+    ),
+    GcovrConfigOption(
+        "exclude_noncode_lines",
+        ["--exclude-noncode-lines"],
+        config="exclude-noncode-lines",
+        group="gcov_options",
+        help="Exclude coverage from lines which seem to be non-code. Default: {default!s}.",
+        action="store_true",
+        const_negate=False,
+    ),
     GcovrConfigOption(
         "exclude_throw_branches",
         ["--exclude-throw-branches"],
@@ -1174,6 +1365,12 @@ GCOVR_CONFIG_OPTIONS = [
         help="Exclude lines that match this regex.",
         type=str,
     ),
+    GcovrConfigOption(
+        "exclude_branches_by_pattern",
+        ["--exclude-branches-by-pattern"],
+        help="Exclude branches that match this regex.",
+        type=str,
+    ),
     GcovrConfigOption(
         "gcov_files",
         ["-g", "--use-gcov-files"],
@@ -1181,17 +1378,46 @@ GCOVR_CONFIG_OPTIONS = [
         help="Use existing gcov files for analysis. Default: {default!s}.",
         action="store_true",
     ),
+    GcovrConfigOption(
+        "gcov_ignore_errors",
+        ["--gcov-ignore-errors"],
+        group="gcov_options",
+        choices=[
+            "all",
+            "no_working_dir_found",
+        ],
+        nargs="?",
+        const="all",
+        default=None,
+        help=(
+            "Ignore errors from invoking GCOV command "
+            "instead of exiting with an error. "
+            "A report will be shown on stderr. "
+            "Default: {default!s}."
+        ),
+        type=str,
+        action="append",
+    ),
     GcovrConfigOption(
         "gcov_ignore_parse_errors",
         ["--gcov-ignore-parse-errors"],
         group="gcov_options",
+        choices=[
+            "all",
+            "negative_hits.warn",
+            "negative_hits.warn_once_per_file",
+        ],
+        nargs="?",
+        const="all",
+        default=None,
         help=(
             "Skip lines with parse errors in GCOV files "
             "instead of exiting with an error. "
             "A report will be shown on stderr. "
             "Default: {default!s}."
         ),
-        action="store_true",
+        type=str,
+        action="append",
     ),
     GcovrConfigOption(
         "objdir",
@@ -1206,6 +1432,7 @@ GCOVR_CONFIG_OPTIONS = [
             "the path from gcc to the gcda file (i.e. gcc's '-o' option), "
             "or the path from the gcda file to gcc's working directory."
         ),
+        type=relative_path,
     ),
     GcovrConfigOption(
         "keep",
