@@ -2,16 +2,17 @@
 
 #  ************************** Copyrights and license ***************************
 #
-# This file is part of gcovr 5.1, a parsing and reporting tool for gcov.
+# This file is part of gcovr 6.0, a parsing and reporting tool for gcov.
 # https://gcovr.com/en/stable
 #
 # _____________________________________________________________________________
 #
-# Copyright (c) 2013-2022 the gcovr authors
+# Copyright (c) 2013-2023 the gcovr authors
 # Copyright (c) 2013 Sandia Corporation.
-# This software is distributed under the BSD License.
 # Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 # the U.S. Government retains certain rights in this software.
+#
+# This software is distributed under the 3-clause BSD License.
 # For more information, see the README.rst file.
 #
 # ****************************************************************************
@@ -23,11 +24,16 @@ import shlex
 import subprocess
 import io
 from threading import Lock
+from typing import Optional
 
-from .utils import search_file, commonpath
+from .utils import search_file, commonpath, is_fs_case_insensitive, fix_case_of_path
 from .workers import locked_directory
-from .coverage import FileCoverage
-from .gcov_parser import parse_metadata, parse_coverage, ParserFlags
+from .gcov_parser import parse_metadata, parse_coverage
+from .coverage import CovData
+from .merging import get_merge_mode_from_options, insert_file_coverage
+from .exclusions import apply_all_exclusions
+from .decision_analysis import DecisionParser
+
 
 logger = logging.getLogger("gcovr")
 
@@ -89,7 +95,9 @@ def find_datafiles(search_path, exclude_dirs):
 #
 # Process a single gcov datafile
 #
-def process_gcov_data(data_fname, covdata, source_fname, options, currdir=None):
+def process_gcov_data(
+    data_fname: str, covdata: CovData, gcda_fname: Optional[str], options, currdir=None
+) -> None:
     with io.open(
         data_fname, "r", encoding=options.source_encoding, errors="replace"
     ) as INPUT:
@@ -98,10 +106,13 @@ def process_gcov_data(data_fname, covdata, source_fname, options, currdir=None):
     # Find the source file
     # TODO: instead of heuristics, use "working directory" if available
     metadata = parse_metadata(lines)
+    source = metadata.get("Source")
+    if source is None:
+        raise RuntimeError("Unexpected value 'None' for metadata 'Source'.")
     fname = guess_source_file_name(
-        metadata["Source"].strip(),
+        source,
         data_fname,
-        source_fname,
+        gcda_fname,
         root_dir=options.root_dir,
         starting_dir=options.starting_dir,
         obj_dir=None if options.objdir is None else os.path.abspath(options.objdir),
@@ -126,59 +137,45 @@ def process_gcov_data(data_fname, covdata, source_fname, options, currdir=None):
 
     key = os.path.normpath(fname)
 
-    parser_flags = ParserFlags.NONE
-    if options.gcov_ignore_parse_errors:
-        parser_flags |= ParserFlags.IGNORE_PARSE_ERRORS
-    if options.exclude_function_lines:
-        parser_flags |= ParserFlags.EXCLUDE_FUNCTION_LINES
-    if options.exclude_internal_functions:
-        parser_flags |= ParserFlags.EXCLUDE_INTERNAL_FUNCTIONS
-    if options.exclude_unreachable_branches:
-        parser_flags |= ParserFlags.EXCLUDE_UNREACHABLE_BRANCHES
-    if options.exclude_throw_branches:
-        parser_flags |= ParserFlags.EXCLUDE_THROW_BRANCHES
-    if options.respect_exclusion_markers:
-        parser_flags |= ParserFlags.RESPECT_EXCLUSION_MARKERS
-    if options.show_decision:
-        parser_flags |= ParserFlags.PARSE_DECISIONS
-
-    coverage = parse_coverage(
+    coverage, source_lines = parse_coverage(
         lines,
         filename=key,
-        exclude_lines_by_pattern=options.exclude_lines_by_pattern,
-        exclude_pattern_prefix=options.exclude_pattern_prefix,
-        flags=parser_flags,
+        ignore_parse_errors=options.gcov_ignore_parse_errors,
     )
-    covdata.setdefault(key, FileCoverage(key)).update(coverage)
+
+    apply_all_exclusions(coverage, lines=source_lines, options=options)
+
+    if options.show_decision:
+        decision_parser = DecisionParser(coverage, source_lines)
+        decision_parser.parse_all_lines()
+
+    insert_file_coverage(covdata, coverage, get_merge_mode_from_options(options))
 
 
 def guess_source_file_name(
-    gcovname,
-    data_fname,
-    source_fname,
-    root_dir,
-    starting_dir,
-    obj_dir,
-    currdir=None,
+    gcovname, data_fname, gcda_fname, root_dir, starting_dir, obj_dir, currdir=None
 ):
     if currdir is None:
         currdir = os.getcwd()
-    if source_fname is None:
+    if gcda_fname is None:
         fname = guess_source_file_name_via_aliases(gcovname, currdir, data_fname)
     else:
         fname = guess_source_file_name_heuristics(
-            gcovname, currdir, root_dir, starting_dir, obj_dir, source_fname
+            gcovname, data_fname, currdir, root_dir, starting_dir, obj_dir, gcda_fname
         )
 
+    if is_fs_case_insensitive():
+        fname = fix_case_of_path(fname)
+
     logger.debug(
         f"Finding source file corresponding to a gcov data file\n"
-        f"  currdir      {currdir}\n"
         f"  gcov_fname   {data_fname}\n"
-        f"  source_fname {source_fname}\n"
+        f"  currdir      {currdir}\n"
         f"  root         {root_dir}\n"
-        # f"  common_dir   {common_dir}\n"
-        # f"  subdir       {subdir}\n"
-        f"  fname        {fname}"
+        f"  starting_dir {starting_dir}\n"
+        f"  obj_dir      {obj_dir}\n"
+        f"  gcda_fname   {gcda_fname}\n"
+        f"  --> fname    {fname}"
     )
 
     return fname
@@ -203,44 +200,51 @@ def guess_source_file_name_via_aliases(gcovname, currdir, data_fname):
 
 
 def guess_source_file_name_heuristics(
-    gcovname, currdir, root_dir, starting_dir, obj_dir, source_fname
+    gcovname, data_fname, currdir, root_dir, starting_dir, obj_dir, gcda_fname
 ):
 
     # gcov writes filenames with '/' path seperators even if the OS
     # separator is different, so we replace it with the correct separator
     gcovname = gcovname.replace("/", os.sep)
 
-    # 0. Try using the current working directory as the source directory
+    # 0. Try using the path to the gcov file
+    fname = os.path.join(os.path.dirname(data_fname), gcovname)
+    if os.path.exists(fname):
+        return fname
+
+    logger.debug("Fallback to heuristic of gcovr 5.1")
+
+    # 1. Try using the current working directory as the source directory
     fname = os.path.join(currdir, gcovname)
     if os.path.exists(fname):
         return fname
 
-    # 1. Try using the path to common prefix with the root_dir as the source directory
+    # 2. Try using the path to common prefix with the root_dir as the source directory
     fname = os.path.join(root_dir, gcovname)
     if os.path.exists(fname):
         return fname
 
-    # 2. Try using the starting directory as the source directory
+    # 3. Try using the starting directory as the source directory
     fname = os.path.join(starting_dir, gcovname)
     if os.path.exists(fname):
         return fname
 
-    # 3. Try using relative path from object dir
+    # 4. Try using relative path from object dir
     if obj_dir is not None:
         fname = os.path.normpath(os.path.join(obj_dir, gcovname))
         if os.path.exists(fname):
             return fname
 
     # Get path of gcda file
-    source_fname_dir = os.path.dirname(source_fname)
+    gcda_fname_dir = os.path.dirname(gcda_fname)
 
-    # 4. Try using the path to the gcda as the source directory
-    fname = os.path.join(source_fname_dir, gcovname)
+    # 5. Try using the path to the gcda as the source directory
+    fname = os.path.join(gcda_fname_dir, gcovname)
     if os.path.exists(fname):
         return os.path.normpath(fname)
 
-    # 5. Try using the path to the gcda file as the source directory, removing the path part from the gcov file
-    fname = os.path.join(source_fname_dir, os.path.basename(gcovname))
+    # 6. Try using the path to the gcda file as the source directory, removing the path part from the gcov file
+    fname = os.path.join(gcda_fname_dir, os.path.basename(gcovname))
     return fname
 
 
@@ -290,7 +294,9 @@ def process_datafile(filename, covdata, options, toerase):
     """
     logger.debug(f"Processing file: {filename}")
 
-    abs_filename = os.path.abspath(filename)
+    abs_filename = os.path.abspath(filename).replace(
+        os.path.sep, "/"
+    )  # gcov requires posix style path
 
     errors = []
 
@@ -318,7 +324,6 @@ def process_datafile(filename, covdata, options, toerase):
             abs_filename,
             covdata,
             options=options,
-            toerase=toerase,
             error=errors.append,
             chdir=wd,
         )
@@ -331,11 +336,19 @@ def process_datafile(filename, covdata, options, toerase):
             return
 
     errors_output = "\n\t".join(errors)
-    logger.warning(
+    errors_output = (
         f"GCOV produced the following errors processing {filename}:\n"
         f"\t{errors_output}\n"
-        "\t(gcovr could not infer a working directory that resolved it.)"
+        "\t(gcovr could not infer a working directory that resolved it.)\n"
+        "To ignore this error use option --gcov-ignore-errors=no_working_dir_found."
     )
+    logger.error(errors_output)
+
+    # Check if error shall be ignored
+    if options.gcov_ignore_errors is None or not any(
+        [v in options.gcov_ignore_errors for v in ["all", "no_working_dir_found"]]
+    ):
+        raise RuntimeError(errors_output)
 
 
 def find_potential_working_directories_via_objdir(abs_filename, objdir, error):
@@ -364,50 +377,62 @@ def find_potential_working_directories_via_objdir(abs_filename, objdir, error):
     return []
 
 
-class gcov:
+class GcovProgram:
     __lock = Lock()
     __cmd = None
     __cmd_split = None
     __default_options = []
     __help_output = None
 
+    class LockContext(object):
+        def __init__(self, lock: Lock):
+            self.lock = lock
+
+        def __enter__(self):
+            self.lock.acquire()
+
+        def __exit__(self, *_):
+            self.lock.release()
+
     def __init__(self, cmd):
-        gcov.__lock.acquire()
-        if gcov.__cmd is None:
-            gcov.__cmd = cmd
-            # If the first element of cmd - the executable name - has embedded spaces
-            # (other than within quotes), it probably includes extra arguments.
-            gcov.__cmd_split = shlex.split(cmd)
-            gcov.__default_options = [
-                "--branch-counts",
-                "--branch-probabilities",
-            ]
-
-            if self.__check_gcov_option("--demangled-names"):
-                gcov.__default_options.append("--demangled-names")
-
-            if self.__check_gcov_option("--hash-filenames"):
-                gcov.__default_options.append("--hash-filenames")
-            elif self.__check_gcov_option("--preserve-paths"):
-                gcov.__default_options.append("--preserve-paths")
+        with GcovProgram.LockContext(GcovProgram.__lock):
+            if GcovProgram.__cmd is None:
+                GcovProgram.__cmd = cmd
+                # If the first element of cmd - the executable name - has embedded spaces
+                # (other than within quotes), it probably includes extra arguments.
+                GcovProgram.__cmd_split = shlex.split(GcovProgram.__cmd)
             else:
-                logger.warning(
-                    "Options '--hash-filenames' and '--preserve-paths' are not "
-                    f"supported by '{cmd}'. Source files with identical file names "
-                    "may result in incorrect coverage."
-                )
-            gcov.__lock.release()
-        else:
-            gcov.__lock.release()
-            assert (
-                gcov.__cmd == cmd
-            ), f"Gcov command must not me changed, expected '{gcov.__cmd}', got '{cmd}'"
+                assert (
+                    GcovProgram.__cmd == cmd
+                ), f"Gcov command must not be changed, expected '{GcovProgram.__cmd}', got '{cmd}'"
+
+    def identify_and_cache_capabilities(self):
+        with GcovProgram.LockContext(GcovProgram.__lock):
+            if not GcovProgram.__default_options:
+                GcovProgram.__default_options = [
+                    "--branch-counts",
+                    "--branch-probabilities",
+                ]
+
+                if self.__check_gcov_option("--demangled-names"):
+                    GcovProgram.__default_options.append("--demangled-names")
+
+                if self.__check_gcov_option("--hash-filenames"):
+                    GcovProgram.__default_options.append("--hash-filenames")
+                elif self.__check_gcov_option("--preserve-paths"):
+                    GcovProgram.__default_options.append("--preserve-paths")
+                else:
+                    logger.warning(
+                        "Options '--hash-filenames' and '--preserve-paths' are not "
+                        f"supported by '{GcovProgram.__cmd}'. Source files with "
+                        "identical file names may result in incorrect coverage."
+                    )
 
     def __get_help_output(self):
-        if gcov.__help_output is None:
-            gcov.__help_output = ""
+        if GcovProgram.__help_output is None:
+            GcovProgram.__help_output = ""
             for help_option in ["--help", "--help-hidden"]:
-                gcov_process = self.run_with_args(
+                gcov_process = self.__get_gcov_process(
                     [help_option],
                     universal_newlines=True,
                 )
@@ -415,12 +440,12 @@ class gcov:
 
                 if not gcov_process.returncode:
                     # gcov execution was successful, help argument is not supported.
-                    gcov.__help_output += out
-            if gcov.__help_output == "":
+                    GcovProgram.__help_output += out
+            if GcovProgram.__help_output == "":
                 # gcov tossed errors: throw exception
                 raise RuntimeError("Error in gcov command line, couldn't get help.")
 
-        return gcov.__help_output
+        return GcovProgram.__help_output
 
     def __check_gcov_option(self, option):
         if option in self.__get_help_output():
@@ -429,9 +454,9 @@ class gcov:
         return False
 
     def get_default_options(self):
-        return gcov.__default_options
+        return GcovProgram.__default_options
 
-    def run_with_args(self, args, **kwargs):
+    def __get_gcov_process(self, args, **kwargs):
         # NB: Currently, we will only parse English output
         env = kwargs.pop("env") if "env" in kwargs else dict(os.environ)
         env["LC_ALL"] = "C"
@@ -439,7 +464,7 @@ class gcov:
 
         if "cwd" not in kwargs:
             kwargs["cwd"] = "."
-        cmd = gcov.__cmd_split + args
+        cmd = GcovProgram.__cmd_split + args
         logger.debug(f"Running gcov: '{' '.join(cmd)}' in '{kwargs['cwd']}'")
 
         return subprocess.Popen(
@@ -451,54 +476,124 @@ class gcov:
             **kwargs,
         )
 
+    def run_with_args(self, args, **kwargs):
+        """Run the gcov program
 
-def run_gcov_and_process_files(abs_filename, covdata, options, error, toerase, chdir):
-    gcov_cmd = gcov(options.gcov_cmd)
+        >>> import platform
+        >>> if platform.system() == "Windows":
+        ...     print("kill not working on Windows")  # doctest: +SKIP
+        ... else:
+        ...     GcovProgram("bash").run_with_args(["-c", "exit 1"])
+        Traceback (most recent call last):
+        ...
+        RuntimeError: GCOV returncode was 1.
+        >>> if platform.system() == "Windows":
+        ...     GcovProgram("bash").run_with_args(["-c", "exit 1"])
+        ... else:
+        ...     print("kill not working on Windows")  # doctest: +SKIP
+        Traceback (most recent call last):
+        ...
+        RuntimeError: GCOV returncode was 4294967295.
+        >>> if platform.system() == "Windows":
+        ...     print("kill not working on Windows")  # doctest: +SKIP
+        ... else:
+        ...     GcovProgram("bash").run_with_args(["-c", "kill $$"])
+        Traceback (most recent call last):
+        ...
+        RuntimeError: GCOV returncode was -15 (exited by signal).
+        >>> if platform.system() == "Windows":
+        ...     GcovProgram("bash").run_with_args(["-c", "kill $$"])
+        ... else:
+        ...     print("kill not working on Windows")  # doctest: +SKIP
+        Traceback (most recent call last):
+        ...
+        RuntimeError: GCOV returncode was 15.
+        """
+        gcov_process = self.__get_gcov_process(args, **kwargs)
+        out, err = gcov_process.communicate()
+        if gcov_process.returncode < 0:
+            raise RuntimeError(
+                f"GCOV returncode was {gcov_process.returncode} (exited by signal)."
+            )
+        elif gcov_process.returncode != 0:
+            raise RuntimeError(f"GCOV returncode was {gcov_process.returncode}.")
 
-    # ATTENTION:
-    # This lock is essential for parallel processing because without
-    # this there can be name collisions for the generated output files.
-    with locked_directory(chdir):
-        out, err = gcov_cmd.run_with_args(
-            [
-                abs_filename,
-                *gcov_cmd.get_default_options(),
-                "--object-directory",
-                os.path.dirname(abs_filename),
-            ],
-            cwd=chdir,
-        ).communicate()
+        return (out, err)
 
-        # find the files that gcov created
-        active_gcov_files, all_gcov_files = select_gcov_files_from_stdout(
-            out,
-            gcov_filter=options.gcov_filter,
-            gcov_exclude=options.gcov_exclude,
-            chdir=chdir,
+
+def run_gcov_and_process_files(abs_filename, covdata, options, error, chdir):
+    fname = None
+    out = None
+    err = None
+    try:
+        gcov_cmd = GcovProgram(options.gcov_cmd)
+        gcov_cmd.identify_and_cache_capabilities()
+
+        # ATTENTION:
+        # This lock is essential for parallel processing because without
+        # this there can be name collisions for the generated output files.
+        with locked_directory(chdir):
+            out, err = gcov_cmd.run_with_args(
+                [
+                    abs_filename,
+                    *gcov_cmd.get_default_options(),
+                    "--object-directory",
+                    os.path.dirname(abs_filename),
+                ],
+                cwd=chdir,
+            )
+
+            # find the files that gcov created
+            active_gcov_files, all_gcov_files = select_gcov_files_from_stdout(
+                out,
+                gcov_filter=options.gcov_filter,
+                gcov_exclude=options.gcov_exclude,
+                chdir=chdir,
+            )
+
+            if unknown_cla_re.search(err):
+                # gcov tossed errors: throw exception
+                raise RuntimeError(f"Error in gcov command line: {err}")
+            elif source_re.search(err):
+                # gcov tossed errors: try the next potential_wd
+                error(err)
+                done = False
+            else:
+                # Process *.gcov files
+                for fname in active_gcov_files:
+                    process_gcov_data(fname, covdata, abs_filename, options)
+                done = True
+
+            if options.keep and done:
+                basename = os.path.basename(abs_filename)
+                for file in active_gcov_files:
+                    dir, filename = os.path.split(file)
+                    os.replace(file, os.path.join(dir, f"{basename}.{filename}"))
+
+            for filepath in (
+                all_gcov_files - active_gcov_files
+                if options.keep and done
+                else all_gcov_files
+            ):
+                if os.path.exists(filepath):
+                    os.remove(filepath)
+
+    except Exception:
+        logger.error(
+            f"Trouble processing {abs_filename!r} with working directory {chdir!r}.\n"
+            f"Stdout of gcov was >>{out}<< End of stdout\n"
+            f"Stderr of gcov was >>{err}<< End of stderr\n"
+            f"Current processed gcov file was {fname!r}.\n"
+            "Use option --verbose to get extended informations."
         )
-
-    if unknown_cla_re.search(err):
-        # gcov tossed errors: throw exception
-        raise RuntimeError("Error in gcov command line: {}".format(err))
-    elif source_re.search(err):
-        # gcov tossed errors: try the next potential_wd
-        error(err)
-        done = False
-    else:
-        # Process *.gcov files
-        for fname in active_gcov_files:
-            process_gcov_data(fname, covdata, abs_filename, options)
-        done = True
-
-    if not options.keep:
-        toerase.update(all_gcov_files)
+        raise
 
     return done
 
 
 def select_gcov_files_from_stdout(out, gcov_filter, gcov_exclude, chdir):
-    active_files = []
-    all_files = []
+    active_files = set([])
+    all_files = set([])
 
     for line in out.splitlines():
         found = output_re.search(line.strip())
@@ -507,7 +602,7 @@ def select_gcov_files_from_stdout(out, gcov_filter, gcov_exclude, chdir):
 
         fname = found.group(1)
         full = os.path.join(chdir, fname)
-        all_files.append(full)
+        all_files.add(full)
 
         filtered, excluded = apply_filter_include_exclude(
             fname, gcov_filter, gcov_exclude
@@ -521,7 +616,7 @@ def select_gcov_files_from_stdout(out, gcov_filter, gcov_exclude, chdir):
             logger.debug(f"Excluding gcov file {fname}")
             continue
 
-        active_files.append(full)
+        active_files.add(full)
 
     return active_files, all_files
 
