@@ -2,16 +2,17 @@
 
 #  ************************** Copyrights and license ***************************
 #
-# This file is part of gcovr 5.1, a parsing and reporting tool for gcov.
+# This file is part of gcovr 6.0, a parsing and reporting tool for gcov.
 # https://gcovr.com/en/stable
 #
 # _____________________________________________________________________________
 #
-# Copyright (c) 2013-2022 the gcovr authors
+# Copyright (c) 2013-2023 the gcovr authors
 # Copyright (c) 2013 Sandia Corporation.
-# This software is distributed under the BSD License.
 # Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 # the U.S. Government retains certain rights in this software.
+#
+# This software is distributed under the 3-clause BSD License.
 # For more information, see the README.rst file.
 #
 # ****************************************************************************
@@ -20,7 +21,7 @@
 Handle parsing of the textual ``.gcov`` file format.
 
 Other modules should only use the following items:
-`parse_metadata()`, `parse_coverage()`, `ParserFlags`, `UnknownLineType`.
+`parse_metadata()`, `parse_coverage()`, `UnknownLineType`.
 
 The behavior of this parser was informed by the following sources:
 
@@ -40,28 +41,36 @@ import logging
 import re
 
 from typing import (
-    List,
+    Any,
     Dict,
+    Iterable,
+    List,
     NamedTuple,
-    Optional,
-    Union,
-    Tuple,
-    Callable,
-    Pattern,
     NoReturn,
+    Optional,
+    Pattern,
+    Tuple,
+    Union,
+)
+
+from .coverage import (
+    BranchCoverage,
+    FileCoverage,
+    FunctionCoverage,
+    CallCoverage,
+    LineCoverage,
+)
+from .merging import (
+    FUNCTION_MAX_LINE_MERGE_OPTIONS,
+    insert_branch_coverage,
+    insert_function_coverage,
+    insert_line_coverage,
+    insert_call_coverage,
 )
 
-from .coverage import FileCoverage
-from .decision_analysis import DecisionParser
 
 logger = logging.getLogger("gcovr")
 
-_EXCLUDE_LINE_FLAG = "_EXCL_"
-_EXCLUDE_LINE_PATTERN_POSTFIX = r"_EXCL_(LINE|START|STOP)"
-
-_C_STYLE_COMMENT_PATTERN = re.compile(r"/\*.*?\*/")
-_CPP_STYLE_COMMENT_PATTERN = re.compile(r"//.*?$")
-
 
 def _line_pattern(pattern: str) -> Pattern[str]:
     """
@@ -78,7 +87,7 @@ def _line_pattern(pattern: str) -> Pattern[str]:
     """
     pattern = pattern.replace(" ", r"[ ]+")
     pattern = pattern.replace("INT", r"[0-9]+")
-    pattern = pattern.replace("VALUE", r"[0-9.]+[%kMGTPEZY]?")
+    pattern = pattern.replace("VALUE", r"(?:NAN %|-?[0-9.]+[%kMGTPEZY]?)")
     return re.compile("^" + pattern + "$")
 
 
@@ -86,7 +95,7 @@ _RE_FUNCTION_LINE = _line_pattern(
     r"function (.*?) called (INT) returned (VALUE) blocks executed (VALUE)"
 )
 _RE_BRANCH_LINE = _line_pattern(
-    r"branch (INT) (?:taken (VALUE)(?: \((\w+)\))?|never executed)"
+    r"branch (INT) (?:taken (VALUE)|never executed)(?: \((\w+)\))?"
 )
 _RE_CALL_LINE = _line_pattern(r"call (INT) (?:returned (VALUE)|never executed)")
 _RE_UNCONDITIONAL_LINE = _line_pattern(
@@ -121,7 +130,7 @@ class _MetadataLine(NamedTuple):
     """A gcov line with metadata: ``-: 0:KEY:VALUE``"""
 
     key: str
-    value: str
+    value: Optional[str]
 
 
 class _BlockLine(NamedTuple):
@@ -171,11 +180,11 @@ class _FunctionLine(NamedTuple):
     """
     A gcov line with function coverage data for the next line.
 
-    ``function NAME called CALLS returned RETURNED blocks executed BLOCKS``
+    ``function NAME called COUNT returned RETURNED blocks executed BLOCKS``
     """
 
     name: str
-    calls: int
+    count: int
     returned: int
     blocks_covered: int
 
@@ -205,7 +214,20 @@ class UnknownLineType(Exception):
         self.line = line
 
 
-def parse_metadata(lines: List[str]) -> Dict[str, str]:
+class NegativeHits(Exception):
+    """Used by `_gather_coverage_from_line()` to signal that a negative count value was found."""
+
+    def __init__(self, line: str) -> None:
+        super().__init__(
+            f"Got negative hit value in gcov line {line!r} caused by a "
+            "bug in gcov tool, see "
+            "https://gcc.gnu.org/bugzilla/show_bug.cgi?id=68080. Use option "
+            "--gcov-ignore-parse-errors with a value of negative_hits.warn, "
+            "or negative_hits.warn_one."
+        )
+
+
+def parse_metadata(lines: List[str]) -> Dict[str, Optional[str]]:
     r"""
     Collect the header/metadata lines from a gcov file.
 
@@ -214,7 +236,19 @@ def parse_metadata(lines: List[str]) -> Dict[str, str]:
     ...   -: 0:Foo:bar
     ...   -: 0:Key:123
     ... '''.splitlines())
-    {'Foo': 'bar', 'Key': '123'}
+    Traceback (most recent call last):
+       ...
+    RuntimeError: Missing key 'Source' in metadata. GCOV data was >>
+      -: 0:Foo:bar
+      -: 0:Key:123<< End of GCOV data
+    >>> parse_metadata('-: 0:Source: file \n -: 0:Foo: bar \n -: 0:Key: 123 '.splitlines())
+    {'Source': 'file', 'Foo': 'bar', 'Key': '123'}
+    >>> parse_metadata('''
+    ...   -: 0:Source:file
+    ...   -: 0:Foo:bar
+    ...   -: 0:Key
+    ... '''.splitlines())
+    {'Source': 'file', 'Foo': 'bar', 'Key': None}
     """
     collected = {}
     for line in lines:
@@ -231,59 +265,24 @@ def parse_metadata(lines: List[str]) -> Dict[str, str]:
         else:
             break  # stop at the first line that is not metadata
 
+    if "Source" not in collected:
+        data = "\n".join(lines)
+        raise RuntimeError(
+            f"Missing key 'Source' in metadata. GCOV data was >>{data}<< End of GCOV data"
+        )
+
     return collected
 
 
-class ParserFlags(enum.Flag):
-    """
-    Flags/toggles that control `parse_coverage()`.
-
-    Multiple values can be combined with binary-or::
-
-        ParserFlags.IGNORE_PASS_ERRORS | ParserFlags.EXCLUDE_THROW_BRANCHES
-    """
-
-    NONE = 0
-    """Default behavior."""
-
-    IGNORE_PARSE_ERRORS = enum.auto()
-    """Whether parse errors shall be converted to warnings."""
-
-    EXCLUDE_FUNCTION_LINES = enum.auto()
-    """Whether coverage shall be ignored on lines that introduce a function."""
-
-    EXCLUDE_INTERNAL_FUNCTIONS = enum.auto()
-    """Whether function coverage for reserved names is ignored."""
-
-    EXCLUDE_UNREACHABLE_BRANCHES = enum.auto()
-    """Whether branch coverage shall be ignored on lines without useful code."""
-
-    EXCLUDE_THROW_BRANCHES = enum.auto()
-    """Whether coverage for exception-only branches shall be ignored."""
-
-    RESPECT_EXCLUSION_MARKERS = enum.auto()
-    """Whether the eclusion markers shall be used."""
-
-    PARSE_DECISIONS = enum.auto()
-    """Whether decision coverage shall be generated."""
-
-
 _LineWithError = Tuple[str, Exception]
 
 
-class _Context(NamedTuple):
-    flags: ParserFlags
-    filename: str
-
-
 def parse_coverage(
     lines: List[str],
     *,
     filename: str,
-    exclude_lines_by_pattern: Optional[str],
-    exclude_pattern_prefix: Optional[str],
-    flags: ParserFlags,
-) -> FileCoverage:
+    ignore_parse_errors: set(),
+) -> Tuple[FileCoverage, List[str]]:
     """
     Extract coverage data from a gcov report.
 
@@ -294,24 +293,18 @@ def parse_coverage(
     Arguments:
         lines: the lines of the file to be parsed (excluding newlines)
         filename: for error reports
-        exclude_lines_by_pattern: string with regex syntax to exclude
-            individual lines
-        exclude_pattern_prefix: string with prefix for _LINE/_START/_STOP markers.
-        flags: various choices for the parser behavior
+        ignore_parse_errors: which errors should be converted to warnings
 
     Returns:
-        the coverage data
+        tuple of the coverage data and the source code lines
 
     Raises:
-        Any exceptions during parsing, unless `ParserFlags.IGNORE_PARSE_ERRORS`
-        is enabled.
+        Any exceptions during parsing, unless ignore_parse_errors is set.
     """
 
-    context = _Context(flags, filename)
-
     lines_with_errors: List[_LineWithError] = []
-
     tokenized_lines: List[Tuple[_Line, str]] = []
+    persistent_states: Dict[str, Any] = {}
     for raw_line in lines:
 
         # empty lines shouldn't occur in reality, but are common in testing
@@ -319,29 +312,22 @@ def parse_coverage(
             continue
 
         try:
-            tokenized_lines.append((_parse_line(raw_line), raw_line))
+            tokenized_lines.append(
+                (
+                    _parse_line(raw_line, ignore_parse_errors, persistent_states),
+                    raw_line,
+                )
+            )
         except Exception as ex:  # pylint: disable=broad-except
             lines_with_errors.append((raw_line, ex))
 
     if (
-        flags & ParserFlags.RESPECT_EXCLUSION_MARKERS
-        or flags & ParserFlags.PARSE_DECISIONS
+        "negative_hits.warn_once_per_file" in persistent_states
+        and persistent_states["negative_hits.warn_once_per_file"] > 1
     ):
-        src_lines = [
-            (line.lineno, line.source_code)
-            for line, _ in tokenized_lines
-            if isinstance(line, _SourceLine)
-        ]
-
-    if flags & ParserFlags.RESPECT_EXCLUSION_MARKERS:
-        line_is_excluded = _find_excluded_ranges(
-            lines=src_lines,
-            warnings=_ExclusionRangeWarnings(filename),
-            exclude_lines_by_pattern=exclude_lines_by_pattern,
-            exclude_pattern_prefix=exclude_pattern_prefix,
+        logger.warning(
+            f"Ignored {persistent_states['negative_hits.warn_once_per_file']} issues overall."
         )
-    else:
-        line_is_excluded = _make_is_in_any_range([])
 
     coverage = FileCoverage(filename)
     state = _ParserState()
@@ -351,8 +337,6 @@ def parse_coverage(
                 state,
                 line,
                 coverage=coverage,
-                line_is_excluded=line_is_excluded,
-                context=context,
             )
         except Exception as ex:  # pylint: disable=broad-except
             lines_with_errors.append((raw_line, ex))
@@ -361,21 +345,37 @@ def parse_coverage(
     # Clean up the final state. This shouldn't happen,
     # but the last line could theoretically contain pending function lines
     for function in state.deferred_functions:
-        _add_coverage_for_function(coverage, state.lineno + 1, function, context)
+        name, count, _, _ = function
+        insert_function_coverage(
+            coverage,
+            FunctionCoverage(name, lineno=state.lineno + 1, count=count),
+            FUNCTION_MAX_LINE_MERGE_OPTIONS,
+        )
 
-    if flags & ParserFlags.PARSE_DECISIONS:
-        decision_parser = DecisionParser(filename, coverage, src_lines)
-        decision_parser.parse_all_lines()
+    _report_lines_with_errors(
+        lines_with_errors,
+        filename=filename,
+        ignore_parse_errors=ignore_parse_errors,
+    )
 
-    _report_lines_with_errors(lines_with_errors, context)
+    src_lines = _reconstruct_source_code(line for line, _ in tokenized_lines)
 
-    return coverage
+    return coverage, src_lines
+
+
+def _reconstruct_source_code(tokens: Iterable[_Line]) -> List[str]:
+    source_token_lines = [line for line in tokens if isinstance(line, _SourceLine)]
+
+    src_lines = [""] * max((line.lineno for line in source_token_lines), default=0)
+    for line in source_token_lines:
+        src_lines[line.lineno - 1] = line.source_code
+
+    return src_lines
 
 
 class _ParserState(NamedTuple):
     deferred_functions: List[_FunctionLine] = []
     lineno: int = 0
-    is_excluded: bool = False
     line_contents: str = ""
     is_recovering: bool = False
 
@@ -385,15 +385,12 @@ def _gather_coverage_from_line(
     line: _Line,
     *,
     coverage: FileCoverage,
-    line_is_excluded: Callable[[int], bool],
-    context: _Context,
 ) -> _ParserState:
     """
     Interpret a Line, updating the FileCoverage, and transitioning ParserState.
 
     The function handles all possible Line variants, and dies otherwise:
-    >>> _gather_coverage_from_line(_ParserState(), "illegal line type",
-    ...     coverage=..., line_is_excluded=..., context=...)
+    >>> _gather_coverage_from_line(_ParserState(), "illegal line type", coverage=...)
     Traceback (most recent call last):
     AssertionError: Unexpected variant: 'illegal line type'
     """
@@ -401,28 +398,26 @@ def _gather_coverage_from_line(
     # pylint: disable=no-else-return  # make life easier for type checkers
 
     if isinstance(line, _SourceLine):
-        lineno = line.lineno
-        is_excluded = line_is_excluded(lineno)
-        noncode, count = _line_noncode_and_count(
-            line,
-            flags=context.flags,
-            is_excluded=is_excluded,
-            is_function=bool(state.deferred_functions),
-        )
+        raw_count, lineno, _, extra_info = line
 
-        if noncode:
-            coverage.line(lineno).noncode = True
-        elif count is not None:
-            coverage.line(lineno).count += count
+        is_noncode = extra_info & _ExtraInfo.NONCODE
+
+        if not is_noncode:
+            insert_line_coverage(coverage, LineCoverage(lineno, count=raw_count))
 
         # handle deferred functions
         for function in state.deferred_functions:
-            _add_coverage_for_function(coverage, line.lineno, function, context)
+            name, count, _, _ = function
+
+            insert_function_coverage(
+                coverage,
+                FunctionCoverage(name, lineno=lineno, count=count),
+                FUNCTION_MAX_LINE_MERGE_OPTIONS,
+            )
 
         return _ParserState(
             lineno=line.lineno,
             line_contents=line.source_code,
-            is_excluded=is_excluded,
         )
 
     elif state.is_recovering:
@@ -434,21 +429,21 @@ def _gather_coverage_from_line(
         return state._replace(deferred_functions=[*state.deferred_functions, line])
 
     elif isinstance(line, _BranchLine):
-        branchno, count, annotation = line
+        branchno, hits, annotation = line
 
-        exclusion_reason = _branch_can_be_excluded(line, state, context.flags)
-        if exclusion_reason:
-            logger.debug(
-                f"Excluding unreachable branch on line {state.lineno} in file {context.filename}: {exclusion_reason}"
+        # line_cov won't exist if it was considered noncode
+        line_cov = coverage.lines.get(state.lineno)
+
+        if line_cov:
+            insert_branch_coverage(
+                line_cov,
+                branchno,
+                BranchCoverage(
+                    count=hits,
+                    fallthrough=(annotation == "fallthrough"),
+                    throw=(annotation == "throw"),
+                ),
             )
-            return state
-
-        branch_cov = coverage.line(state.lineno).branch(branchno)
-        branch_cov.count += count
-        if annotation == "fallthrough":
-            branch_cov.fallthrough = True
-        if annotation == "throw":
-            branch_cov.throw = True
 
         return state
 
@@ -461,7 +456,21 @@ def _gather_coverage_from_line(
         return state
 
     # ignore unused line types, such as specialization sections
-    elif isinstance(line, (_CallLine, _UnconditionalLine, _BlockLine)):
+    elif isinstance(line, _CallLine):
+        callno, returned = line
+        line_cov = coverage.lines[state.lineno]  # must already exist
+
+        insert_call_coverage(
+            line_cov,
+            CallCoverage(
+                callno=callno,
+                covered=(returned > 0),
+            ),
+        )
+
+        return state
+
+    elif isinstance(line, (_UnconditionalLine, _BlockLine)):
         return state
 
     else:
@@ -474,7 +483,10 @@ def _assert_never(never: NoReturn) -> NoReturn:
 
 
 def _report_lines_with_errors(
-    lines_with_errors: List[_LineWithError], context: _Context
+    lines_with_errors: List[_LineWithError],
+    *,
+    filename: str,
+    ignore_parse_errors: set,
 ) -> None:
     """Log warnings and potentially re-throw exceptions"""
 
@@ -486,7 +498,7 @@ def _report_lines_with_errors(
 
     lines_output = "\n\t  ".join(lines)
     logger.warning(
-        f"Unrecognized GCOV output for {context.filename}\n"
+        f"Unrecognized GCOV output for {filename}\n"
         f"\t  {lines_output}\n"
         "\tThis is indicative of a gcov output parse error.\n"
         "\tPlease report this to the gcovr developers\n"
@@ -496,7 +508,7 @@ def _report_lines_with_errors(
     for ex in errors:
         logger.warning(f"Exception during parsing:\n\t{type(ex).__name__}: {ex}")
 
-    if context.flags & ParserFlags.IGNORE_PARSE_ERRORS:
+    if ignore_parse_errors is not None and "all" in ignore_parse_errors:
         return
 
     logger.error(
@@ -509,97 +521,9 @@ def _report_lines_with_errors(
     raise errors[0]  # guaranteed to have at least one exception
 
 
-def _line_noncode_and_count(
-    line: _SourceLine, *, flags: ParserFlags, is_excluded: bool, is_function: bool
-) -> Tuple[bool, Optional[int]]:
-    """
-    Some reports like JSON are sensitive to which lines are excluded,
-    so keep this convoluted logic for now.
-
-    The count field is only meaningful if not(noncode).
-    """
-
-    raw_count, _, source_code, extra_info = line
-
-    if flags & ParserFlags.EXCLUDE_FUNCTION_LINES and is_function:
-        return True, None
-
-    if is_excluded:
-        return True, None
-
-    if extra_info & _ExtraInfo.NONCODE:
-        if _is_non_code(source_code):
-            return True, None
-        return False, None  # completely ignore this line
-
-    if raw_count == 0 and _is_non_code(source_code):
-        return True, None
-
-    return False, raw_count
-
-
-def _function_can_be_excluded(name: str, flags: ParserFlags) -> bool:
-    # special names for construction/destruction of static objects will be ignored
-    if flags & ParserFlags.EXCLUDE_INTERNAL_FUNCTIONS:
-        if name.startswith("__") or name.startswith("_GLOBAL__sub_I_"):
-            return True
-
-    return False
-
-
-def _branch_can_be_excluded(
-    branch: _BranchLine, state: _ParserState, flags: ParserFlags
-) -> Optional[str]:
-    if state.is_excluded:
-        return "marked with exclude pattern"
-
-    if flags & ParserFlags.EXCLUDE_UNREACHABLE_BRANCHES:
-        if not _line_can_contain_branches(state.line_contents):
-            return "detected as compiler-generated code"
-
-    if flags & ParserFlags.EXCLUDE_THROW_BRANCHES:
-        if branch.annotation == "throw":
-            return "detected as exception-only code"
-
-    return None
-
-
-def _line_can_contain_branches(code: str) -> bool:
-    """
-    False if the line looks empty except for braces.
-
-    >>> _line_can_contain_branches('} // end something')
-    False
-    >>> _line_can_contain_branches('foo();')
-    True
-    """
-
-    code = _CPP_STYLE_COMMENT_PATTERN.sub("", code)
-    code = _C_STYLE_COMMENT_PATTERN.sub("", code)
-    code = code.strip().replace(" ", "")
-    return code not in ["", "{", "}", "{}"]
-
-
-def _add_coverage_for_function(
-    coverage: FileCoverage,
-    lineno: int,
-    function: _FunctionLine,
-    context: _Context,
-) -> None:
-    name, calls, _, _ = function
-
-    if _function_can_be_excluded(name, context.flags):
-        logger.debug(
-            f"Ignoring Symbol {name} in line {lineno} in file {context.filename}"
-        )
-        return
-
-    function_cov = coverage.function(name)
-    function_cov.lineno = lineno
-    function_cov.count += calls
-
-
-def _parse_line(line: str) -> _Line:
+def _parse_line(
+    line: str, ignore_parse_errors: set = (), persistent_states: dict = {}
+) -> _Line:
     """
     Categorize/parse individual lines without further processing.
 
@@ -630,12 +554,16 @@ def _parse_line(line: str) -> _Line:
     _BranchLine(branchno=3, hits=0, annotation=None)
     >>> _parse_line('branch 3 taken 123')
     _BranchLine(branchno=3, hits=123, annotation=None)
+    >>> _parse_line('branch 3 taken -1', ("negative_hits.warn",))
+    _BranchLine(branchno=3, hits=0, annotation=None)
     >>> _parse_line('branch 7 taken 3% (fallthrough)')
     _BranchLine(branchno=7, hits=1, annotation='fallthrough')
     >>> _parse_line('branch 17 taken 99% (throw)')
     _BranchLine(branchno=17, hits=1, annotation='throw')
     >>> _parse_line('branch  0 never executed')
     _BranchLine(branchno=0, hits=0, annotation=None)
+    >>> _parse_line('branch  0 never executed (fallthrough)')
+    _BranchLine(branchno=0, hits=0, annotation='fallthrough')
     >>> _parse_line('branch 2 with some unknown format')
     Traceback (most recent call last):
     gcovr.gcov_parser.UnknownLineType: branch 2 with some unknown format
@@ -660,7 +588,7 @@ def _parse_line(line: str) -> _Line:
 
     Example: can parse function tags:
     >>> _parse_line('function foo called 2 returned 95% blocks executed 85%')
-    _FunctionLine(name='foo', calls=2, returned=1, blocks_covered=1)
+    _FunctionLine(name='foo', count=2, returned=1, blocks_covered=1)
     >>> _parse_line('function foo with some unknown format')
     Traceback (most recent call last):
     gcovr.gcov_parser.UnknownLineType: function foo with some unknown format
@@ -679,7 +607,6 @@ def _parse_line(line: str) -> _Line:
     Traceback (most recent call last):
     gcovr.gcov_parser.UnknownLineType: :
 
-
     Example: can parse block line:
     >>> _parse_line('     1: 32-block  0')
     _BlockLine(hits=1, lineno=32, blockno=0, extra_info=NONE)
@@ -698,7 +625,7 @@ def _parse_line(line: str) -> _Line:
     """
     # pylint: disable=too-many-branches
 
-    tag = _parse_tag_line(line)
+    tag = _parse_tag_line(line, ignore_parse_errors, persistent_states)
     if tag is not None:
         return tag
 
@@ -721,8 +648,12 @@ def _parse_line(line: str) -> _Line:
 
         # METADATA (key, value)
         if count_str == "-" and lineno == "0":
-            key, value = source_code.split(":", 1)
-            return _MetadataLine(key, value)
+            if ":" in source_code:
+                key, value = source_code.split(":", 1)
+                return _MetadataLine(key, value.strip())
+            else:
+                # Add a syntethic metadata with no value
+                return _MetadataLine(source_code, None)
 
         if count_str == "-":
             count = 0
@@ -777,7 +708,9 @@ def _parse_line(line: str) -> _Line:
     raise UnknownLineType(line)
 
 
-def _parse_tag_line(line: str) -> Optional[_Line]:
+def _parse_tag_line(
+    line: str, ignore_parse_errors: set = (), persistent_states: dict = {}
+) -> Optional[_Line]:
     """A tag line is any gcov line that starts in the first column."""
     # pylint: disable=too-many-return-statements
 
@@ -799,6 +732,19 @@ def _parse_tag_line(line: str) -> Optional[_Line]:
         if match is not None:
             branch_id, taken_str, annotation = match.groups()
             hits = 0 if taken_str is None else _int_from_gcov_unit(taken_str)
+            if hits < 0:
+                if ignore_parse_errors is not None and (
+                    "negative_hits.warn" in ignore_parse_errors
+                    or "negative_hits.warn_once_per_file" in ignore_parse_errors
+                ):
+                    if "negative_hits.warn_once_per_file" not in persistent_states:
+                        persistent_states["negative_hits.warn_once_per_file"] = 0
+                        logger.warning(f"Ignoring negative hits in line {line!r}.")
+                    persistent_states["negative_hits.warn_once_per_file"] += 1
+                else:
+                    raise NegativeHits(line)
+                hits = 0
+
             return _BranchLine(int(branch_id), hits, annotation)
 
     # CALL
@@ -832,10 +778,10 @@ def _parse_tag_line(line: str) -> Optional[_Line]:
     if line.startswith("function "):
         match = _RE_FUNCTION_LINE.match(line)
         if match is not None:
-            name, calls, returns, blocks = match.groups()
+            name, count, returns, blocks = match.groups()
             return _FunctionLine(
                 name,
-                _int_from_gcov_unit(calls),
+                _int_from_gcov_unit(count),
                 _int_from_gcov_unit(returns),
                 _int_from_gcov_unit(blocks),
             )
@@ -849,238 +795,6 @@ def _parse_tag_line(line: str) -> Optional[_Line]:
     return None
 
 
-class _ExclusionRangeWarnings:
-    r"""
-    Log warnings related to exclusion marker processing.
-
-    Example:
-    >>> source = '''\
-    ...  1: 1: some code
-    ...  1: 2: foo // LCOV_EXCL_STOP
-    ...  1: 3: bar // GCOVR_EXCL_START
-    ...  1: 4: bar // GCOVR_EXCL_LINE
-    ...  1: 5: baz // GCOV_EXCL_STOP
-    ...  1: 6: "GCOVR_EXCL_START"
-    ... '''
-    >>> caplog = getfixture('caplog')
-    >>> caplog.clear()
-    >>> _ = parse_coverage(  # doctest: +NORMALIZE_WHITESPACE
-    ...     source.splitlines(), filename='example.cpp',
-    ...     flags=ParserFlags.RESPECT_EXCLUSION_MARKERS, exclude_lines_by_pattern=None,
-    ...     exclude_pattern_prefix='[GL]COVR?')
-    >>> for message in caplog.record_tuples:
-    ...     print(f"{message[1]}: {message[2]}")
-    30: mismatched coverage exclusion flags.
-              LCOV_EXCL_STOP found on line 2 without corresponding LCOV_EXCL_START, when processing example.cpp.
-    30: GCOVR_EXCL_LINE found on line 4 in excluded region started on line 3, when processing example.cpp.
-    30: GCOVR_EXCL_START found on line 3 was terminated by GCOV_EXCL_STOP on line 5, when processing example.cpp.
-    30: The coverage exclusion region start flag GCOVR_EXCL_START
-              on line 6 did not have corresponding GCOVR_EXCL_STOP flag
-              in file example.cpp.
-    """
-
-    def __init__(self, filename: str) -> None:
-        self.filename = filename
-
-    def mismatched_start_stop(
-        self, start_lineno: int, start: str, stop_lineno: int, stop: str
-    ) -> None:
-        """warn that start/stop region markers don't match"""
-        logger.warning(
-            f"{start} found on line {start_lineno} "
-            f"was terminated by {stop} on line {stop_lineno}, "
-            f"when processing {self.filename}."
-        )
-
-    def stop_without_start(self, lineno: int, expected_start: str, stop: str) -> None:
-        """warn that a region was ended without corresponding start marker"""
-        logger.warning(
-            "mismatched coverage exclusion flags.\n"
-            f"          {stop} found on line {lineno} without corresponding {expected_start}, "
-            f"when processing {self.filename}."
-        )
-
-    def start_without_stop(self, lineno: int, start: str, expected_stop: str) -> None:
-        """warn that a region was started but not closed"""
-        logger.warning(
-            f"The coverage exclusion region start flag {start}\n"
-            f"          on line {lineno} did not have corresponding {expected_stop} flag\n"
-            f"          in file {self.filename}."
-        )
-
-    def line_after_start(self, lineno: int, start: str, start_lineno: str) -> None:
-        """warn that a region was started but an excluded line was found"""
-        logger.warning(
-            f"{start} found on line {lineno} in excluded region started on line {start_lineno}, "
-            f"when processing {self.filename}."
-        )
-
-
-def _find_excluded_ranges(
-    lines: List[Tuple[int, str]],
-    *,
-    warnings: _ExclusionRangeWarnings,
-    exclude_lines_by_pattern: Optional[str] = None,
-    exclude_pattern_prefix: str,
-) -> Callable[[int], bool]:
-    """
-    Scan through all lines to find line ranges covered by exclusion markers.
-
-    Example:
-    >>> lines = [(11, '//PREFIX_EXCL_LINE'), (13, '//IGNORE'), (15, '//PREFIX_EXCL_START'), (18, '//PREFIX_EXCL_STOP')]
-    >>> exclude = _find_excluded_ranges(
-    ...     lines, warnings=..., exclude_lines_by_pattern = '.*IGNORE', exclude_pattern_prefix='PREFIX')
-    >>> [lineno for lineno in range(20) if exclude(lineno)]
-    [11, 13, 15, 16, 17]
-    """
-    exclude_lines_by_pattern_regex = None
-    if exclude_lines_by_pattern:
-        exclude_lines_by_pattern_regex = re.compile(exclude_lines_by_pattern)
-
-    # possibly overlapping half-open ranges of lines that are excluded
-    exclude_line_ranges: List[Tuple[int, int]] = []
-
-    exclusion_stack = []
-    for lineno, code in lines:
-        if _EXCLUDE_LINE_FLAG in code:
-            # process the exclusion marker
-            #
-            # header is a marker name like LCOV or GCOVR
-            #
-            # START flags are added to the exlusion stack
-            # STOP flags remove a marker from the exclusion stack
-            excl_line_pattern = re.compile(
-                "(" + exclude_pattern_prefix + ")" + _EXCLUDE_LINE_PATTERN_POSTFIX
-            )
-            for header, flag in excl_line_pattern.findall(code):
-
-                if flag == "LINE":
-                    if exclusion_stack:
-                        warnings.line_after_start(
-                            lineno, f"{header}_EXCL_LINE", exclusion_stack[-1][1]
-                        )
-                    else:
-                        exclude_line_ranges.append((lineno, lineno + 1))
-
-                if flag == "START":
-                    exclusion_stack.append((header, lineno))
-
-                elif flag == "STOP":
-                    if not exclusion_stack:
-                        warnings.stop_without_start(
-                            lineno, f"{header}_EXCL_START", f"{header}_EXCL_STOP"
-                        )
-                        continue
-
-                    start_header, start_lineno = exclusion_stack.pop()
-                    if header != start_header:
-                        warnings.mismatched_start_stop(
-                            start_lineno,
-                            f"{start_header}_EXCL_START",
-                            lineno,
-                            f"{header}_EXCL_STOP",
-                        )
-
-                    exclude_line_ranges.append((start_lineno, lineno))
-
-                else:  # pragma: no cover
-                    pass
-
-        if exclude_lines_by_pattern_regex:
-            if exclude_lines_by_pattern_regex.match(code):
-                exclude_line_ranges.append((lineno, lineno + 1))
-
-    for header, lineno in exclusion_stack:
-        warnings.start_without_stop(
-            lineno, f"{header}_EXCL_START", f"{header}_EXCL_STOP"
-        )
-
-    return _make_is_in_any_range(exclude_line_ranges)
-
-
-def _make_is_in_any_range(ranges: List[Tuple[int, int]]) -> Callable[[int], bool]:
-    """
-    Create a function to check whether an input is in any range.
-
-    This function should provide reasonable performance
-    if queries are mostly made in ascending order.
-
-    Example:
-    >>> select = _make_is_in_any_range([(3,4), (5,7)])
-    >>> select(0)
-    False
-    >>> select(6)
-    True
-    >>> [x for x in range(10) if select(x)]
-    [3, 5, 6]
-    """
-
-    # values are likely queried in ascending order,
-    # allowing the search to start with the first possible range
-    ranges = sorted(ranges)
-    hint_value = 0
-    hint_index = 0
-
-    def is_in_any_range(value: int) -> bool:
-        nonlocal hint_value, hint_index
-
-        # if the heuristic failed, restart search from the beginning
-        if value < hint_value:
-            hint_index = 0
-
-        hint_value = value
-
-        for i in range(hint_index, len(ranges)):
-            start, end = ranges[i]
-            hint_index = i
-
-            # stop as soon as a too-large range is seen
-            if value < start:
-                break
-
-            if start <= value < end:
-                return True
-        else:
-            hint_index = len(ranges)
-
-        return False
-
-    return is_in_any_range
-
-
-def _is_non_code(code: str) -> bool:
-    """
-    Check for patterns that indicate that this line doesn't contain useful code.
-
-    Examples:
-    >>> _is_non_code('  // some comment!')
-    True
-    >>> _is_non_code('  /* some comment! */')
-    True
-    >>> _is_non_code('} else {')  # could be easily made detectable
-    False
-    >>> _is_non_code('}else{')
-    False
-    >>> _is_non_code('else')
-    True
-    >>> _is_non_code('{')
-    True
-    >>> _is_non_code('/* some comment */ {')
-    True
-    >>> _is_non_code('}')
-    True
-    >>> _is_non_code('} // some code')
-    True
-    >>> _is_non_code('return {};')
-    False
-    """
-
-    code = _CPP_STYLE_COMMENT_PATTERN.sub("", code)
-    code = _C_STYLE_COMMENT_PATTERN.sub("", code)
-    code = code.strip()
-    return len(code) == 0 or code in ["{", "}", "else"]
-
-
 def _int_from_gcov_unit(formatted: str) -> int:
     """
     Try to reverse gcov's number formatting.
@@ -1106,8 +820,10 @@ def _int_from_gcov_unit(formatted: str) -> int:
     Examples:
     >>> _int_from_gcov_unit('123')
     123
-    >>> [_int_from_gcov_unit(value) for value in ('17.2%', '0%')]
-    [1, 0]
+    >>> _int_from_gcov_unit('-1.2k')
+    -1200
+    >>> [_int_from_gcov_unit(value) for value in ('NAN %', '17.2%', '0%')]
+    [0, 1, 0]
     >>> [_int_from_gcov_unit(value) for value in ('1.7k', '0.5G')]
     [1700, 500000000]
     """
