@@ -2,12 +2,12 @@
 
 #  ************************** Copyrights and license ***************************
 #
-# This file is part of gcovr 5.2, a parsing and reporting tool for gcov.
+# This file is part of gcovr 6.0, a parsing and reporting tool for gcov.
 # https://gcovr.com/en/stable
 #
 # _____________________________________________________________________________
 #
-# Copyright (c) 2013-2022 the gcovr authors
+# Copyright (c) 2013-2023 the gcovr authors
 # Copyright (c) 2013 Sandia Corporation.
 # Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 # the U.S. Government retains certain rights in this software.
@@ -20,14 +20,14 @@
 import csv
 from typing import Tuple, Optional
 
-from ..utils import sort_coverage, presentable_filename, open_text_for_writing
-from ..coverage import CovData, CoverageStat, SummarizedStats
+from ..utils import presentable_filename, open_text_for_writing
+from ..coverage import CovData, CoverageStat, SummarizedStats, sort_coverage
 
 
 def print_csv_report(covdata: CovData, output_file, options):
     """produce gcovr csv report"""
 
-    with open_text_for_writing(output_file, "coverage.csv") as fh:
+    with open_text_for_writing(output_file, "coverage.csv", newline="") as fh:
         keys = sort_coverage(
             covdata,
             show_branch=options.show_branch,
