@@ -2,12 +2,12 @@
 
 #  ************************** Copyrights and license ***************************
 #
-# This file is part of gcovr 5.2, a parsing and reporting tool for gcov.
+# This file is part of gcovr 6.0, a parsing and reporting tool for gcov.
 # https://gcovr.com/en/stable
 #
 # _____________________________________________________________________________
 #
-# Copyright (c) 2013-2022 the gcovr authors
+# Copyright (c) 2013-2023 the gcovr authors
 # Copyright (c) 2013 Sandia Corporation.
 # Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 # the U.S. Government retains certain rights in this software.
@@ -20,11 +20,11 @@
 from typing import Iterable, Tuple
 
 from ..utils import (
-    sort_coverage,
+    force_unix_separator,
     presentable_filename,
     open_text_for_writing,
 )
-from ..coverage import CovData, CoverageStat, FileCoverage
+from ..coverage import CovData, CoverageStat, FileCoverage, sort_coverage
 
 # Widths of the various columns
 COL_FILE_WIDTH = 40
@@ -43,7 +43,7 @@ def print_text_report(covdata: CovData, output_file, options):
         fh.write("-" * LINE_WIDTH + "\n")
         fh.write("GCC Code Coverage Report".center(LINE_WIDTH).rstrip() + "\n")
         # fh.write(" " * 27 + "GCC Code Coverage Report\n")
-        fh.write("Directory: " + options.root.replace("\\", "/") + "\n")
+        fh.write("Directory: " + force_unix_separator(options.root) + "\n")
 
         fh.write("-" * LINE_WIDTH + "\n")
         title_total = "Branches" if options.show_branch else "Lines"
